from flockai.models.drones.autopilot_controlled_drone import AutopilotControlledDrone
from flockai.models.drones.keyboard_controller_drone import KeyboardControlledDrone
from flockai.models.probes.flockai_probe import FlockAIProbe
from flockai.models.devices.device_enums import EnableableDevice, NonEnableableDevice
import time


class KeyboardMavic2DJI(KeyboardControlledDrone):
    """
    A Keyboard Controlled Mavic2DJI
    """

    def __init__(self, devices, probe: FlockAIProbe=None, model=None):
        super().__init__(devices)
        self.probe = probe
        self._activate_probes()
        self.model = model

    def _activate_probes(self):
        if self.probe is not None:
            self.probe.activate()
            print(f'{self.probe.get_name()} probe activated')

    def blink_led_lights(self, led_devices: list):
        """
        Blink led lights every 2 seconds
        :param time: The simulation time. ie self.getTime()
        :return:
        """
        f_led_devices = [self.devices[led]['device'] for led in led_devices]
        led_on = (self.getTime() // 1) % 2 == 1

        for f_led in f_led_devices:
            f_led.set(led_on)
            led_on = not led_on

    def get_processing_energy(self):
        probe_alive_time = self.probe.get_metric('ProbeAliveTimeMetric')
        probe_cpu_time = self.probe.get_metric('ProcessCpuTimeMetric')
        probe_io_time = self.probe.get_metric('ProcessIOTimeMetric')

        cpu_time = probe_cpu_time.get_val()
        flight_time = probe_alive_time.get_val()
        t = probe_io_time.get_val()
        io_time = 0 if t is None else t

        return self.energy_model.processing_energy.calculate(cpu_time_active=cpu_time, flight_time=flight_time, io_time=io_time)

    def get_communication_energy(self, comm_transmit_time, comm_receive_time):
        probe_alive_time = self.probe.get_metric('ProbeAliveTimeMetric')
        flight_time = probe_alive_time.get_val()
        return self.energy_model.communication_energy.calculate(transmit_time=comm_transmit_time, receive_time=comm_receive_time, idle_time=flight_time-(comm_transmit_time + comm_receive_time))

    def run(self):
        # Wait a second before starting
        while self.step(self.basic_time_step) != -1:
            if self.getTime() > 1:
                break

        led_devices = []
        emitter_devices = []
        receiver_devices = []
        for device_name in self.devices:
            if self.devices[device_name]['type'] == NonEnableableDevice.LED:
                led_devices.append(device_name)
            if self.devices[device_name]['type'] == NonEnableableDevice.EMITTER:
                emitter_devices.append(device_name)
            if self.devices[device_name]['type'] == EnableableDevice.RECEIVER:
                receiver_devices.append(device_name)

        comm_transmit_time = 0
        comm_receive_time = 0

        while self.step(self.basic_time_step) != -1:
            self.blink_led_lights(led_devices)
            self.actuate()

            # Drone sends data specified by the user to the base station if ML is not running on board
            # TODO:
            t1 = time.time()
            self.send_msg(msg="test", emitter_devices=emitter_devices)
            t2 = time.time()
            comm_transmit_time += t2 - t1
            t1 = time.time()
            received_messages = self.receive_msgs(receiver_devices=receiver_devices)
            t2 = time.time()
            comm_receive_time += t2 - t1

            # Execute prediction pipeline step
            if self.model is not None:
                prediction = self.model.predict()
                # if prediction is not None:
                #     print(prediction)

            # Get energy values on each cycle
            energy = {**self.get_processing_energy(),
                      **self.get_communication_energy(comm_transmit_time=comm_transmit_time, comm_receive_time=comm_receive_time)}
            energy['total_energy'] = energy['e_proc'] + energy['e_comm']

            # Update battery values
            self.battery.remaining_energy_percentage = 1.0 - (energy['total_energy'] / self.battery.energy_capacity)

            if self.getTime() % 5.0 != 0.0:  # get access to controller functions
                continue

            # Update log data every 5 minutes with energy values
            print(energy, f'{self.battery.remaining_energy_percentage * 100}% battery remaining')


class AutopilotMavic2DJI(AutopilotControlledDrone):
    def __init__(self, devices, probe: FlockAIProbe=None, model=None):
        super().__init__(devices)

    def send_msg(self, msg, emitter_devices: list):
        """
        Sends a message from the attached emitter device
        :param emitter_devices: The emitter devices to send the message
        :param msg: The message to send
        :return: None
        """
        total_time = 0

        for emitter in emitter_devices:
            if emitter not in self.devices:
                print(f"The specified emitter device is not found: {emitter}")
                return 0
            f_emitter = self.devices[emitter]['device']

        return total_time

    def run(self):
        # Wait a second before starting
        while self.step(self.basic_time_step) != -1:
            if self.getTime() > 1:
                break

        emitter_devices = []
        for device_name in self.devices:
            if self.devices[device_name]['type'] == NonEnableableDevice.EMITTER:
                emitter_devices.append(device_name)

        while self.step(self.basic_time_step) != -1:
            # For now just actuate
            self.actuate()



from flockai.interfaces.energy import IEnergy


class MotorEnergy(IEnergy):
    """
    Energy used to operate motors. Ie. Electrical Motors, Propellers...
    """
    def __init__(self):
        self.e_takeoff = 0  # energy attributed to the drone lifted in the air till it enters a hovering state
        self.e_hover = 0  # energy attributed to keeping the drone in the air and is dependent to the air density and weight of the drone
        self.e_move = 0  # the kinetic energy required to set a drone in motion from a hovering state

        # Need to be defined
        self.p_hover = 0

    def calculate(self, hovering_time):
        self.e_hover = self.p_hover * hovering_time
        return {"e_hover": self.e_hover,
                "e_motor_total": self.e_hover + self.e_move + self.e_takeoff}

    def set(self, *args, **kwargs):
        pass


class CommunicationEnergy(IEnergy):
    """
    Energy used to operate communication devices
    """

    def __init__(self):
        self.e_transmit = 0  # energy attributed to transmit data
        self.t_transmit = 0  # time attributed to transmit data
        self.e_receive = 0   # energy attributed to receive data
        self.t_receive = 0   # energy attributed to receive data
        self.e_idle = 0      # energy attributed when idling
        self.t_idle = 0      # idle time

        # Need to define these
        self.p_transmit = 0
        self.p_receive = 0
        self.p_idle = 0

    def calculate(self, transmit_time, receive_time, idle_time):
        self.e_transmit = self.p_transmit * transmit_time
        self.e_receive = self.p_receive * receive_time
        self.e_idle = self.p_idle * idle_time

        return {"e_transmit": self.e_transmit,
                "e_receive": self.e_receive,
                "e_idle": self.e_idle,
                "e_comm": self.e_transmit + self.e_receive + self.e_idle}

    def set(self, *args, **kwargs):
        pass


class ProcessingEnergy(IEnergy):
    """
    Energy used for processing
    """
    def __init__(self):
        self.e_cpu_active = 0      # processor power in compute state
        self.t_cpu_active = 0      # cpu time
        self.e_cpu_io = 0       # processor power in I/O state
        self.t_cpu_io = 0       # I/O time
        self.e_cpu_peripheral = 0  # processor power driving peripheral sensing devices
        self.t_cpu_peripheral = 0  # peripheral usage time
        self.e_cpu_idle = 0     # processor power when idling
        self.t_cpu_idle = 0     # idle time
        self.e_fc = 0           # flight controller energy

        # Need to be defined
        self.p_fc = 0
        self.p_cpu_active = 0
        self.p_cpu_idle = 0
        self.p_cpu_io = 0
        self.p_cpu_peripheral = 0

    def calculate(self, cpu_time_active, flight_time, io_time):
        self.e_cpu_active = self.p_cpu_active * cpu_time_active
        self.e_cpu_idle = self.p_cpu_idle * (flight_time - cpu_time_active)
        self.e_fc = self.p_fc * (flight_time / 10)
        self.e_cpu_io = self.p_cpu_io * io_time

        return {"e_cpu_active": self.e_cpu_active,
                "e_cpu_idle": self.e_cpu_idle,
                "e_fc": self.e_fc,
                "e_cpu_io": self.e_cpu_io,
                "e_proc": self.e_cpu_active + self.e_cpu_idle + self.e_fc + self.e_cpu_io}

    def set(self, *args, **kwargs):
        pass


class Energy:
    # Comm
    P_COMM = 8.4
    P_IDLE = 4.2

    # Proc
    DJI_A3_FC = 8
    RASPBERRY_PI_4B_IDLE = 4
    RASPBERRY_PI_4B_ACTIVE = 8

    # Motor
    DJI_P_HOVER = 95.02

    def __init__(self):
        self.motor_energy = MotorEnergy()
        self.communication_energy = CommunicationEnergy()
        self.processing_energy = ProcessingEnergy()




import random
import time

from flockai.PyCatascopia.Probe import Probe
from flockai.PyCatascopia.Metrics import SimpleMetric, CounterMetric, DiffMetric, TimerMetric


class ExampleProbe(Probe):

    def __init__(self, name = "ExampleProbe", periodicity = 5):
        super(ExampleProbe, self).__init__(name, periodicity)

        self.myMetric1 = SimpleMetric('myMetric1', '%', 'random double between 0 and 10', 0, 10)
        self.myMetric2 = SimpleMetric('myMetric2', '#', 'random int between 0 and 1000', 0, 1000, higherIsBetter=False)
        self.myMetric3 = CounterMetric('myMetric3', '#', 'counter incrementing by 1 and reseting at 20', maxVal=20)
        self.myMetric4 = DiffMetric('myMetric4', '#', 'scaled difference from previous val')
        self.myMetric5 = TimerMetric('myMetric5', maxVal=10)

        self.add_metric(self.myMetric1)
        self.add_metric(self.myMetric2)
        self.add_metric(self.myMetric3)
        self.add_metric(self.myMetric4)
        self.add_metric(self.myMetric5)

    def get_desc(self):
        return "ExampleProbe collects some dummy metrics..."

    def collect(self):
        self.myMetric5.timer_reset_and_start()

        d = random.uniform(0, 10)
        i = random.randint(0, 1000)

        self.myMetric1.set_val(d)
        self.myMetric2.set_val(i)
        self.myMetric3.inc()
        self.myMetric4.update(i)

        time.sleep(d)
        self.myMetric5.timer_end()

def main():
    p = ExampleProbe()
    p.set_debugmode(True)
    p.set_logging()
    p.activate()

if __name__ == "__main__":
    main()



class Battery:
    DJI_CHARGE_CAPACITY = 3850  # mAh
    DJI_VOLTAGE = 15.4  # V
    DJI_ENERGY_CAPACITY = 59.29 * 3600  # W
    DJI_MAX_FLIGHT_TIME = 31 * 60  # seconds
    DJI_MAX_HOVERING_TIME = 29 * 60  # seconds
    DJI_SAFE_LANDING_PERCENTAGE = 0.1  # percentage

    def __init__(self):
        self.charge_capacity = 0
        self.voltage = 0
        self.energy_capacity = 0
        self.max_flight_time = 0
        self.max_hover_time = 0
        self.safe_landing_percentage = 0
        self.remaining_energy_percentage = 1.0

